from pathlib import Path
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
import textwrap
import sys


colors_for_models = {
    "AS-RCEC_TaiESM1_r1i1p1f1": "#0000FF",
    "AWI_AWI-CM-1-1-MR_r1i1p1f1": "#00FF00",
    "CCCma_CanESM5_r1i1p1f1": "#FF0000",
    "CCCR-IITM_IITM-ESM_r1i1p1f1": "#EEDD00",
    "CMCC_CMCC-CM2-SR5_r1i1p1f1": "#FF00FF",
    "CMCC_CMCC-ESM2_r1i1p1f1": "#00FFFF",
    "CNRM-CERFACS_CNRM-CM6-1-HR_r1i1p1f2": "#000088",
    "CNRM-CERFACS_CNRM-CM6-1_r1i1p1f2": "#008800",
    "CNRM-CERFACS_CNRM-ESM2-1_r1i1p1f2": "#880000",
    "CSIRO-ARCCSS_ACCESS-CM2_r1i1p1f1": "#888800",
    "EC-Earth-Consortium_EC-Earth3-CC_r1i1p1f1": "#880088",
    "INM_INM-CM4-8_r1i1p1f1": "#008888",
    "INM_INM-CM5-0_r1i1p1f1": "#8888FF",
    "IPSL_IPSL-CM6A-LR_r1i1p1f1": "#88FF88",
    "KIOST_KIOST-ESM_r1i1p1f1": "#FF8888",
    "MIROC_MIROC6_r1i1p1f1": "#EEDD88",
    "MIROC_MIROC-ES2L_r1i1p1f2": "#FF88FF",
    "MOHC_HadGEM3-GC31-LL_r1i1p1f3": "#88FFFF",
    "MOHC_HadGEM3-GC31-MM_r1i1p1f3": "#0088FF",
    "MOHC_UKESM1-0-LL_r1i1p1f2": "#00FF88",
    "MPI-M_MPI-ESM1-2-LR_r1i1p1f1": "#88FF00",
    "MRI_MRI-ESM2-0_r1i1p1f1": "#FF8800",
    "NCAR_CESM2-WACCM_r1i1p1f1": "#8800FF",
    "NCC_NorESM2-MM_r1i1p1f1": "#FF0088",
    "NIMS-KMA_KACE-1-0-G_r1i1p1f1": "#000000",
    "NOAA-GFDL_GFDL-ESM4_r1i1p1f1": "#666666",
    "NUIST_NESM3_r1i1p1f1": "#BBBBBB"
}


variables = dict()

class Variable:
    def __init__(self, name, unit, caption=None, manipulator=None):
        self.name = name
        self.unit = unit
        
        self.caption = caption if (caption != None) else name.replace("-", " ")
        self.manipulator = manipulator if (manipulator != None) else lambda x : x
        
        variables[name] = self

class Axis(Variable):
    def __init__(self, variable, area, season):
        super().__init__(variable.name, variable.unit, variable.caption, variable.manipulator)
        
        self.area = area
        self.season = season

Variable("cloud-cover", "%")
Variable("cooling-degree-days", "°C·day/year")
Variable("evaporation", "mm/year", manipulator=(lambda x : 365.25 * x))
Variable("frost-days", "per year")
Variable("heating-degree-days", "°C·day/year")
Variable("humidity", "%", manipulator=(lambda x : 100 * x))
Variable("precipitation", "mm/year", manipulator=(lambda x : 365.25 * x))
Variable("snowfall", "mm/year", manipulator=(lambda x : 365.25 * x))
Variable("soil-moisture", "kg/m²")
Variable("surface-downwelling-longwave-radiation", "W/m²", "surface ↓ longwave radiation")
Variable("surface-downwelling-shortwave-radiation", "W/m²", "surface ↓ shortwave radiation")
Variable("temperature", "°C")
Variable("wind-speed", "m/s")


class GraphData():
    
    def __init__(self, name_x, name_y, area_x, area_y, season_x, season_y, avg_range, id):
        self.id = id
        
        self.x = Axis(variables[name_x], area_x, season_x)
        self.y = Axis(variables[name_y], area_y, season_y)
        
        path_x = f"data/{area_x}/{name_x}/{season_x}.csv"
        path_y = f"data/{area_y}/{name_y}/{season_y}.csv"
        df_x = pd.read_csv(path_x, skiprows=1)
        df_y = pd.read_csv(path_y, skiprows=1)
        self.points = self.points_by_model(df_x, df_y)
        
        self.avg_points = self.moving_average(avg_range)
    
    def points_by_model(self, df_x, df_y):
        points = []
        
        for model in df_x:
            if (model == "Year") or (not model in df_y):
                continue
            
            points_x = []
            points_y = []
            points.append((model, points_x, points_y))
            
            model_x = df_x[model]
            model_y = df_y[model]
            
            for i in range(2025-1850, min(len(model_x), len(model_y))):
                variable_x = model_x[i]
                variable_y = model_y[i]
                
                try:
                    variable_x = float(variable_x)
                    variable_y = float(variable_y)
                    points_x.append(self.x.manipulator(variable_x))
                    points_y.append(self.y.manipulator(variable_y))
                except:
                    continue
        
        return points

    def moving_average(self, avg_range):
        average_points = []
        
        for model, points_x, points_y in self.points:
            avg_x = []
            avg_y = []
            average_points.append((model, avg_x, avg_y))
            
            for i in range(len(points_x) - avg_range + 1):
                avg_x.append(np.average(points_x[i:i+avg_range]))
                avg_y.append(np.average(points_y[i:i+avg_range]))
        
        return average_points


def draw_graph(ax, points):
    points_x = []
    points_y = []
        
    # Scatter points
    for i in range(len(points)):
        model_points = points[i]
        color = colors_for_models[model_points[0]]
        ax.scatter(model_points[1], model_points[2], s=3, c=color)
        
        points_x.extend(model_points[1])
        points_y.extend(model_points[2])
    
    # Draw trend line of scattered points
    coefficients = np.polyfit(points_x, points_y, 1)
    trend_line = np.poly1d(coefficients)
    ax.plot(points_x, trend_line(points_x), color="black", linewidth=1)


def stylize_var(axis):    
    if (axis.area == "global"):
        return f"{axis.season} global {axis.caption}"
    
    return f"{axis.season} {axis.caption} in {axis.area.capitalize()}"


def stylize_axis(axis):
    text = stylize_var(axis)
    text = text[0].upper() + text[1:]
    text = f"Δ {text}"
    
    wrapped = textwrap.wrap(text, 35)
    wrapped[-1] += f" ({axis.unit})"
    
    return "\n".join(wrapped)


def graph_subtitle(graph):
    stylized_y = stylize_var(graph.y)
    stylized_x = stylize_var(graph.x)
    
    text = f"Figure {graph.id}. Change in {stylized_y} relative to {stylized_x}."
    wrapped = textwrap.wrap(text, 57)
    return "\n".join(wrapped)


def create_subplot(ax, graph):
    ax.set_xlabel(stylize_axis(graph.x))
    ax.set_ylabel(stylize_axis(graph.y))
    
    subtitle = graph_subtitle(graph)
    ax.text(0.5, -0.30, subtitle, transform=ax.transAxes, ha="center", fontweight="light")

    draw_graph(ax, graph.avg_points)


def subplot_with_global(ax, name_y, season_y, avg_range, id):
    graph_data = GraphData("temperature", name_y, "global", "estonia", "annual", season_y, avg_range, id)
    create_subplot(ax, graph_data)


def suptitle(season, average):
    if (season == "annual"):
        season = ""
    else:
        season += " "
    
    return f"Changes in Estonia compared to changes in global temperature relative to 1850-1900.\n" \
           f"CMIP6 {season}predictions of 2025-2100 ({average}-year averages). SSP5-8.5 scenario."


def temperature_graphs(season, average):
    figure, ((ax11, ax12, ax13), (ax21, ax22, ax23)) = plt.subplots(2, 3)
    figure.suptitle(suptitle(season, average))
    
    subplot_with_global(ax11, "temperature", season, average, "1a")
    subplot_with_global(ax12, "surface-downwelling-longwave-radiation", season, average, "1b")
    subplot_with_global(ax13, "surface-downwelling-shortwave-radiation", season, average, "1c")
    subplot_with_global(ax21, "frost-days", season, average, "1d")
    subplot_with_global(ax22, "cooling-degree-days", season, average, "1e")
    subplot_with_global(ax23, "heating-degree-days", season, average, "1f")


def precipitation_graphs(season, average):
    figure, ((ax11, ax12), (ax21, ax22)) = plt.subplots(2, 2)
    figure.suptitle(suptitle(season, average))
    
    subplot_with_global(ax11, "precipitation", season, average, "2a")
    subplot_with_global(ax12, "snowfall", season, average, "2b")
    subplot_with_global(ax21, "humidity", season, average, "2c")
    subplot_with_global(ax22, "evaporation", season, average, "2d")


def wind_graphs(season, average):
    figure, ((ax11, ax12), (ax21, ax22)) = plt.subplots(2, 2)
    figure.suptitle(suptitle(season, average))
    
    subplot_with_global(ax11, "wind-speed", season, average, "3a")
    subplot_with_global(ax12, "cloud-cover", season, average, "3b")
    subplot_with_global(ax21, "surface-downwelling-longwave-radiation", season, average, "1b")
    subplot_with_global(ax22, "surface-downwelling-shortwave-radiation", season, average, "1c")


def main():
    
    season = sys.argv[2] if len(sys.argv) > 2 else "annual"
    
    if (sys.argv[1] == "temperature"): temperature_graphs(season, 5)
    if (sys.argv[1] == "precipitation"): precipitation_graphs(season, 5)
    if (sys.argv[1] == "wind"): wind_graphs(season, 5)
    
    plt.subplots_adjust(left=0.09, right=0.91, 
                        top=0.9, bottom=0.13, 
                        wspace=0.3, hspace=0.4)

    plt.show()

main()