from pathlib import Path
import matplotlib.pyplot as plt
import pandas as pd
import math
import sys

displayed_percentiles = [90, 75, 50, 25, 10]

def is_float(value):
    try:
        value = float(value)
        return True
    except:
        return False
    
def find_percentile(values, percentile):
    sorted_values = values[:]
    sorted_values.sort()
    
    length = len(values) - 1
    index = percentile / 100 * length
    index1 = math.floor(index)
    index2 = math.ceil(index)
    weight1 = 1 - (index % 1)
    weight2 = (index % 1)

    return round((weight1 * sorted_values[index1]) + (weight2 * sorted_values[index2]), 2)

class Container:
    def __init__(self, value):
        self.value = value

class GraphData():
    
    class TempsPerModel():
        def __init__(self, years, temperatures):
            self.years = years
            self.temperatures = temperatures
    
    def get_temperatures_per_model(self, df):
        temperatures_per_model = []
        years = df["Year"]
        
        for header, column in df.items():
            if (header == "Year"): continue
            
            avail_years = []
            avail_temps = []
            for i in range(df.shape[0]):
                temperature = column[i]
                if not is_float(temperature): continue
                avail_years.append(years[i])
                avail_temps.append(float(temperature))
            
            temperatures_per_model.append(GraphData.TempsPerModel(avail_years, avail_temps))
    
        return temperatures_per_model

    
    def get_percentiles_per_year(self, df):
        percentiles_per_year = dict()
        
        for row in df.iloc:
            year = row[0]
            temperature_cells = row[1:]
        
            temperatures = []
            for cell in temperature_cells:
                if (is_float(cell)):
                    temperatures.append(float(cell))
            
            percentiles_per_year[year] = [(p, find_percentile(temperatures, p)) for p in displayed_percentiles]
        
        return percentiles_per_year
    
    
    def get_medians_per_year(self, df):
        medians_per_year = []
        
        for row in df.iloc:
            temperature_cells = row[1:]
        
            temperatures = []
            for cell in temperature_cells:
                if (is_float(cell)):
                    temperatures.append(float(cell))
            
            medians_per_year.append(find_percentile(temperatures, 50))
        
        return medians_per_year
    
    
    def get_graph_corners(self, df):
        years = df["Year"]
        
        min_temp = math.inf
        max_temp = -math.inf
        
        for header, column in df.items():
            if (header == "Year"): continue
            
            for i in range(df.shape[0]):
                temperature = column[i]
                if not is_float(temperature): continue
                min_temp = min(min_temp, float(temperature))
                max_temp = max(max_temp, float(temperature))
                    
        return (years[0], min_temp, years[len(years)-1], max_temp)
    
    def __init__(self, df):
        self.temps_per_model = self.get_temperatures_per_model(df)
        self.percentiles_per_year = self.get_percentiles_per_year(df)
        self.medians_per_year = self.get_medians_per_year(df)
        self.years = df["Year"]
        (self.xmin, self.ymin, self.xmax, self.ymax) = self.get_graph_corners(df)


def get_hover_text(year, data):
    percentiles_per_year = data.percentiles_per_year
    year = round(year)
    
    if not year in percentiles_per_year:
        return None
    
    text = f"{year}:"
    
    for percentiles in percentiles_per_year[year]:
        temp_text = f"{percentiles[1]}"
        if (not temp_text.startswith("-")):
            temp_text = "+" + temp_text
        
        text += f"\n{percentiles[0]}th percentile: {temp_text} °C"
    
    return text


def draw_mouse_line(event, ax, data, mouse_line):
    if (mouse_line.value != None):
        mouse_line.value[0].remove()
    
    year = round(event.xdata)
    year = max(data.xmin, year)
    year = min(data.xmax, year)
    
    mouse_line.value = ax.plot([year, year], [data.ymin, data.ymax], linewidth="1", color="#999999ff")

def draw_hover_text(event, annotation, data):
    year = event.xdata
    year = max(data.xmin, year)
    year = min(data.xmax, year)
    
    annotation.xy = (year, event.ydata)
    annotation.set_text(get_hover_text(year, data))
    
    if (year > 2024):
        annotation.xyann = (-145, -50)
    else:
        annotation.xyann = (15, -50)


def on_mouse_move(event, figure, ax, annotation, data, mouse_line):
    if not event.inaxes == ax:
        return
    
    draw_hover_text(event, annotation, data)
    draw_mouse_line(event, ax, data, mouse_line)
        
    figure.canvas.draw_idle()


def create_hover_annotation(figure, ax, data):
    annotation = ax.annotate(
        text = "",
        xy = (0, 0),
        xytext = (0, 0),
        textcoords = "offset points",
        bbox = dict(boxstyle="round,pad=0.5", edgecolor="#ffffff00", facecolor="#dddddddd"),
        fontsize = 10,
        visible = False
    )
    
    mouse_line = Container(None)
    
    annotation.set_visible(True)
    figure.canvas.mpl_connect("motion_notify_event", lambda event: on_mouse_move(event, figure, ax, annotation, data, mouse_line))
    
    return annotation


def draw_graph(dataframe, type):
    data = GraphData(dataframe)

    figure, (ax) = plt.subplots(1)
    type_stylized = type[0].upper() + type[1:].lower()
    ax.set_title(f"Climate projections in Estonia - {type_stylized}")
    ax.set_ylabel("Temperature (°C)")
    ax.set_xlabel("Year")
    
    # Draw temperatures of different models
    for temperatures in data.temps_per_model:
        ax.plot(temperatures.years, temperatures.temperatures, linestyle="dashed", linewidth="0.5", color="#ee000055")
    
    # Draw median temperatures
    ax.plot(data.years, data.medians_per_year, linewidth="2", color="#cc0000")
    
    # Line at current year
    ax.plot([2024, 2024], [data.ymin, data.ymax], linewidth="1", color="#666666ff")
    
    create_hover_annotation(figure, ax, data)
    
    plt.show()


def main():
    type = "annual"
    if (len(sys.argv) >= 2):
        type = sys.argv[1].lower()
    
    path = f"data/mean-temperatures/{type}.csv"
    if not Path(path).exists():
        print(f"Could not find projections for '{type}'")
    else:    
        dataframe = pd.read_csv(path, skiprows=1)
        draw_graph(dataframe, type)

main()